open Term 

let rec subst x t = function 
    | TVar y when x = y -> t
    | TAdd (e1, e2) -> TAdd (subst x t e1, subst x t e2) 
    | TSub (e1, e2) -> TSub (subst x t e1, subst x t e2) 
    | TMul (e1, e2) -> TMul (subst x t e1, subst x t e2) 
    | TEq (e1, e2) -> TEq (subst x t e1, subst x t e2) 
    | TLt (e1, e2) -> TLt (subst x t e1, subst x t e2) 
    | TApp (e1, e2) -> TApp (subst x t e1, subst x t e2) 
    | TFun (y, e) when y<> x -> TFun (y, subst x t e) 
    | TIf (e1, e2, e3) -> TIf (subst x t e1, subst x t e2, subst x t e3) 
    | TPair (e1, e2) -> TPair (subst x t e1, subst x t e2) 
    | TLet (y, e1, e2) -> TLet (y, subst x t e1, if x = y then e2 else subst x t e2) 
    | TLetTuple (y, z, e1, e2) -> TLetTuple (y, z, subst x t e1, if x=y || x=z then e2 else subst x t e2) 
    | TFix (f, y, e) when f <> x && y <> x -> TFix (f, y, subst x t e) 
    | TSeq (e1, e2) ->  TSeq (subst x t e1, subst x t e2) 
    | TAss (e1, e2) -> TAss (subst x t e1, subst x t e2)
    | TDeref e -> TDeref (subst x t e)
    | TRef e -> TRef (subst x t e)
    | e -> e

let subst_value x v = subst x (TCst v)

let rec eval = function 
    | TVar x -> failwith (Printf.sprintf "eval: open term (%s)" x)
    | TCst v -> v
    | TAdd (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VInt n1, VInt n2 -> VInt (n1 + n2)
          | _ -> failwith "eval: type error")
    | TMul (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VInt n1, VInt n2 -> VInt (n1 * n2)
          | _ -> failwith "eval: type error")
    | TSub (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VInt n1, VInt n2 -> VInt (n1 - n2)
          | _ -> failwith "eval: type error")
    | TEq (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VInt n1, VInt n2 -> VBool (n1 = n2)
          | VBool b1, VBool b2 -> VBool (b1 = b2)
          | _ -> failwith "eval: type error")
    | TLt (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VInt n1, VInt n2 -> VBool (n1 < n2)
          | _ -> failwith "eval: type error")
    | TApp (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VFun f, v -> f v
          | _ -> failwith "eval: type error")
    | TFun (x, e) -> 
         VFun (fun v -> eval (subst_value x v e))
    | TIf (e1, e2, e3) -> 
        (match eval e1 with 
          | VBool b -> if b then eval e2 else eval e3
          | _ -> failwith "eval: type error")
    | TPair (e1, e2) -> VPair (eval e1, eval e2)
    | TLet (x, e1, e2) -> eval (subst_value x (eval e1) e2)
    | TLetTuple (x,y,e1,e2) -> 
        (match eval e1 with 
          | VPair (v1, v2) -> eval (subst_value x v1 (subst_value y v2 e2))
          | _ -> failwith "eval: type error")
    | TFix (f, x, e) as fix -> 
        VFun (fun v -> eval (subst_value f (eval fix) (subst_value x v e)))
    | TRef t -> VRef (ref (eval t))
    | TDeref t -> 
        (match eval t with 
          | VRef r -> !r
          | _ -> failwith "eval: type error")
    | TAss (e1, e2) -> 
        (match eval e1, eval e2 with 
          | VRef r, v -> r := v; VUnit
          | _ -> failwith "eval: type error")
    | TSeq (e1, e2) -> ignore (eval e1); eval e2 


